#include "darioblob.h"

#include "cv.h"
#include "cvaux.h"
//#include "cxcore.h"
#include "highgui.h"
//#include "cvcam.h"
#include <stdio.h>
#include <algorithm>

//#define DEBUGGAMI


DvBlob creaDvBlob(float X, float Y, float W, float H, bool sel) {
	DvBlob risultato;
	risultato.x = X;
	risultato.y = Y;
	risultato.w = W;
	risultato.h = H;
	risultato.ID = -1;
	risultato.selected = sel;

	return (risultato);
}

bool cmp( DvBlob a, DvBlob b ) {
	return a.ID > b.ID; 
}


// ++++++++
// BLOBLIST
// ++++++++
bloblist::bloblist(int id) {
	lista = new std::vector<DvBlob>;
	totale = 0;
	lowerID = id;
}

bloblist::~bloblist() {
	lista->clear();
	delete (lista);
}

int bloblist::addBlob(DvBlob* d, bool autoid) {

	int risultato = -1;
	
	// Trovo l'ID pi basso
	if (autoid) {
		(*d).ID = lowerID;
		risultato = lowerID;
		lista->push_back(*d);
		lowerID ++;
		
	} else {
		lista->push_back(*d);
		risultato = d->ID;
		calcolaLowerID();	
	}

	totale ++;

	return risultato;

	
}


void bloblist::delBlob(int quale) {
	if ((quale <= totale) && (quale >= 0)) {
		lista->erase(lista->begin() + quale);
		totale --;
		calcolaLowerID();
	}
}

void bloblist::delBlobById(int id) {
	std::vector<DvBlob>::iterator iter;
	iter = lista->begin();
	bool fatto = false;
	while (!fatto) {

		if (iter == lista->end()) {
			fatto = true;
		}

		if ((*iter).ID == id) {
			lista->erase(iter);
			fatto = true;
			totale --;
			calcolaLowerID();		
		}

		iter ++;
	}

	

}

DvBlob* bloblist::getBlob(int quale) {

	if ((quale < totale) && (quale >= 0)) {
		return &(lista->at(quale));
	} else {
		return NULL;
	}

}

DvBlob* bloblist::getBlobById(int id) {
	DvBlob* risultato = NULL;

	std::vector<DvBlob>::iterator iter;

	bool fatto = false;

	iter = lista->begin();

	while(!fatto) {
		if ((iter) == lista->end()) {
			fatto = true;
		}

		if ((*iter).ID == id) {
			risultato = &(*iter);
			fatto = true;
		}

		iter ++;
	}

	return (risultato);

}

void bloblist::svuotaLista() {
	//printf("Svuoto la lista\n");
	lista->clear();
	totale = 0;
	lowerID = 0;
	//printf("Lista svuotata\n");
}

void bloblist::tuttiFalse() {
	std::vector<DvBlob>::iterator iter;

	for (iter = lista->begin(); iter != lista->end(); iter ++) {
		(*iter).selected = false;
	}
}

void bloblist::delNonSelected() {

	std::vector<DvBlob>* nuovaLista = new std::vector<DvBlob>;
	std::vector<DvBlob>::iterator iter;

	for (iter = lista->begin(); iter != lista->end(); iter++) {
		if ((*iter).selected == true) {
			nuovaLista->push_back((*iter));
		}
	}

	lista->clear();
	delete (lista);
	lista = nuovaLista;

}

void bloblist::calcolaLowerID() {

	// Sorto la lista
	sort( lista->begin(), lista->end(), cmp );     

	std::vector<DvBlob>::reverse_iterator iter;
	iter = lista->rbegin();
	lowerID = (*iter).ID + 1;

}


int bloblist::getLowerId() {
	
	return lowerID;
}


// +++++++++
// DARIOBLOB
// +++++++++

darioblob::darioblob(const IplImage* img, void (*cb)(char*)) {
	sfondo = NULL;

	IplImage* copia = cvCloneImage(img);
	sfondo = copia;

	valSoglia = 5;
	maxSoglia = 255;
	valErode = 5;
	valDilate = 10;

	// Creo la lista dei blob
	listaBlob = new bloblist();

	// Assegno il callback
	if (cb != NULL) {
		callbackEventi = cb;
	}

}

darioblob::~darioblob() {
	if (sfondo != NULL) cvReleaseImage(&sfondo);

	delete (listaBlob);
}

void darioblob::updateSfondo(const IplImage* img) {
	if (img != NULL) {
		if (sfondo != NULL) cvReleaseImage(&sfondo);
		sfondo = cvCloneImage(img);
	}

}

IplImage* darioblob::process(const IplImage* img) {


	// +++++++++++++++++++++++++
	// ELIMINAZIONE DELLO SFONDO
	// +++++++++++++++++++++++++

	
	IplImage* risultato = cvCloneImage(img);
	IplImage* scalaDiGrigi = cvCreateImage(cvGetSize(img), 8, 1);
	IplImage* soglia = cvCreateImage(cvGetSize(img), 8, 1);


	// Sottraggo lo sfondo al frame, e salvo nel risultato
	cvAbsDiff(img, sfondo, risultato);

	// Converto il risultato in scala di grigi
	cvCvtColor(risultato, scalaDiGrigi, CV_RGB2GRAY);

	// Applico la soglia
	cvThreshold(scalaDiGrigi, soglia, valSoglia, maxSoglia, CV_THRESH_BINARY);

	// Erosione dell'immagine per filtrare
	cvErode(soglia, soglia, NULL, valErode);

	// Dilate dell'immagine per inciccire le forme
	cvDilate(soglia, soglia, NULL, valDilate);

	// Rilascio le immagini che non servono pi
	cvReleaseImage(&risultato);
	cvReleaseImage(&scalaDiGrigi);


	// ++++++++++++++++
	// RICERCA DEI BLOB
	// ++++++++++++++++
        CvSeq*      cnts = NULL;
        CvSeq*      cnt = NULL; // Serve per il for qui sotto
		CvMemStorage*   m_pMem = cvCreateMemStorage();;
		IplImage* copia = cvCloneImage(soglia);
		bloblist* blobTrovati = new bloblist();

        int i = 0;

		cvClearMemStorage(m_pMem);

		// Trovo tutti i contorni

		cvFindContours(copia, m_pMem, &cnts, sizeof(CvContour), CV_RETR_EXTERNAL);

		// Analizzo tutti i contorni

		for (cnt = cnts; cnt; cnt = cnt->h_next) {

			DvBlob  NewBlob;
            // image moments
            double      M00,X,Y,XX,YY;
            CvMoments   m;
            CvRect      r = ((CvContour*)cnt)->rect;
            CvMat       mat;
            if(r.height < 3 || r.width < 3) continue;
            cvMoments( cvGetSubRect(copia,&mat,r), &m, 0 );
            M00 = cvGetSpatialMoment( &m, 0, 0 );
            if(M00 <= 0 ) continue;
            X = cvGetSpatialMoment( &m, 1, 0 )/M00;
            Y = cvGetSpatialMoment( &m, 0, 1 )/M00;
            XX = (cvGetSpatialMoment( &m, 2, 0 )/M00) - X*X;
            YY = (cvGetSpatialMoment( &m, 0, 2 )/M00) - Y*Y;
            NewBlob = creaDvBlob(r.x+(float)X, r.y+(float)Y, (float)(4*sqrt(XX)), (float)(4*sqrt(YY)));
			blobTrovati->addBlob(&NewBlob);
			//i ++;

		}

		
		// Rilascio la memoria utilizzata
		if(m_pMem != NULL)cvReleaseMemStorage(&m_pMem);


		// ++++++++++++
		// TRACCIAMENTO
		// ++++++++++++

		// Se non ho trovato blob in questo frame, devo cancellare quelli gi esistenti
		if (blobTrovati->getTotale() == 0) {
			// Cancello la lista dei blob esistenti

			for (int i = 0; i < listaBlob->getTotale(); i ++) {
				char messaggio[100];
				sprintf(messaggio, "Il blob %d esce", listaBlob->getBlob(i)->ID);
				pushEvent(messaggio);
			}

			listaBlob->svuotaLista();


		} else {
			//listaBlob->tuttiFalse();
			bloblist* listaNuova = new bloblist(listaBlob->getLowerId());

			// Ho gi dei blob: devo accoppiarli
			// Trovo il vicino
			// Se NON c', aggiungo quel blob alla lista

			
			// Ordino la lista dei blob per ID
			//listaBlob->getLowerId();

			// Scansiono tutti i nuovi blob
			for (int i = 0; i < blobTrovati->getTotale(); i ++) {

				DvBlob* esaminando = blobTrovati->getBlob(i); // Il blob in esame ora
				int chi = getVicino(esaminando); // Trovo il suo vicino
				
				if (chi != -1) {
					// Ho trovato il vicino
					DvBlob* vicino = listaBlob->getBlobById(chi); // Ecco il vicino

					if (vicino->selected == false) {
						
						// Creo un nuovo blob
						DvBlob nuovo = creaDvBlob(esaminando->x, esaminando->y, esaminando->h, esaminando->w);
						// Sistemo l'ID
						nuovo.ID = vicino->ID;
						nuovo.selected = false;

						//printf("vicino->ID = %d\n", vicino->ID);


						// Aggiungo sto blob alla lista nuova
						listaNuova->addBlob(&nuovo, false);

						// Imposto il vicino come selected
						vicino->selected = true;

						// Imposto il blob in blobTrovati come selected
						esaminando->selected = true;

						//char messaggio[100];
						//sprintf(messaggio, "Il blob %d si muove", nuovo.ID);
						//pushEvent(messaggio);
					} else {
						DvBlob nuovo = creaDvBlob(esaminando->x, esaminando->y, esaminando->w, esaminando->h);
						int valore = listaNuova->addBlob(&nuovo, true);
						esaminando->selected = true;
						char messaggio[100];
						sprintf(messaggio, "Nuovo Blob: %d", valore);
						pushEvent(messaggio);

					}
				} else {
					// Non ho trovato nessun vicino: aggiungo questo blob e via
					esaminando->selected = true;
					DvBlob novello = creaDvBlob(esaminando->x, esaminando->y, esaminando->w, esaminando->h);
					novello.selected = false;
					int id = listaNuova->addBlob(&novello, true);


					char messaggio[100];
					sprintf(messaggio, "Nuovo Blob: %d", id);
					pushEvent(messaggio);

				}
			}


			// Mando l'evento dei blob morenti
			for (int i = 0; i < listaBlob->getTotale(); i ++) {
				DvBlob* asen = listaBlob->getBlob(i);
				if (listaBlob->getBlob(i)->selected == false) {
					char messaggio[100];
					sprintf(messaggio, "Il blob %d esce", asen->ID);
					pushEvent(messaggio);
				}
			}


			delete (listaBlob);
			listaBlob = listaNuova;


			//listaBlob->tuttiFalse();
			//printf("Alla fine, listaNuova contiene %d blob\n", listaNuova->getTotale());
		}

		int bt = blobTrovati->getTotale();
		int lb = listaBlob->getTotale();
		if (bt != lb) {
			//printf("Differenza: trovati = %d, totale = %d\n", bt, lb);
		}

		delete (blobTrovati);
	return (soglia);

}

bloblist* darioblob::getLista() {
	return listaBlob;
}


int darioblob::getVicino(DvBlob* blob) {
	int risultato = -1;

	if (blob == NULL) return (risultato);

	DvBlob* esaminando = NULL;
	double DistBest = -1;

	for (int i = 0; i < listaBlob->getTotale(); i ++) {
		esaminando = listaBlob->getBlob(i);

		double Dist = -1;
		double dx = fabs(blob->x - esaminando->x);
		double dy = fabs(blob->y - esaminando->y);

		if ( (dx > 2*(blob->w)) || (dy > 2*(blob->h)) ) continue;

		Dist = sqrt(dx*dx + dy*dy);
		if (Dist < DistBest || risultato == -1) {
			DistBest = Dist;
			risultato = esaminando->ID;
		}
	}

	return risultato;

}


void darioblob::pushEvent(char* stringa) {
	if (callbackEventi != NULL) {
		callbackEventi(stringa);
	}
}

